#include "common.h"

#include <arpa/inet.h>
#include <assert.h>
#include <stdint.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>

#ifdef HAS_NUMA
#include <numa.h>
#include <numaif.h>
#endif

#include <infiniband/verbs.h>
#include <rdma/rdma_cma.h>

#include <endian.h>
#include <assert.h>

int init_common_channel(struct common *common) {
    common->cm_channel = rdma_create_event_channel();
    if (!common->cm_channel) {
        rdma_error("rdma_create_event_channel() failed");
        return 1;
    }
    
    return 0;
}

int allocate_variables(struct common *common, struct common_setup_config *config) {
    common->buf = calloc(config->variables, sizeof (uint32_t*));
    if (!common->buf)
        return 1;

    if (config->randomize_variable_home) {
        for (unsigned i = 0; i < config->variables; ++i) {
            int ret = posix_memalign((void **)&common->buf[i], getpagesize(), sizeof(uint32_t));
            if (ret) {
                rdma_error("posix_memalign() returned %d\n", ret);
                return 1;
            }

#ifdef HAS_NUMA
            int node = rand() % (numa_max_node() + 1);
            unsigned long nodemask = 1 << node;
            ret = mbind(common->buf[i], getpagesize(), MPOL_BIND, &nodemask,
                    32, MPOL_MF_MOVE | MPOL_MF_STRICT);
            if (ret) {
                rdma_error("Failed to bind to node %d, errno: %d \n", i, errno);
                return 1;
            }
#endif
        }
    } else {
        common->buf[0] = calloc(config->variables, sizeof (uint32_t));
        if (!common->buf[0])
            return 1;

        for (unsigned i = 1; i < config->variables; ++i) {
            common->buf[i] = &common->buf[0][i];
        }
    }

    if (config->variables_initialization_value) {
        for (unsigned i = 0; i < config->variables; ++i)
            *common->buf[i] = config->variables_initialization_value;
    }

    return 0;
}

int init_common(struct common *common, struct common_setup_config* config) {
    struct ibv_qp_init_attr qp_attr = {};
    int ret;

    common->variables = config->variables;

    /* Create verbs objects now that we know which device to use */

    /* Allocate protection domain; each pd can be used to create queue pair,
     * register memory region, etc.
     * Each pd is a protection of a group of objects,
     * it means you can't use these objects between different pd.
     */
    common->pd = ibv_alloc_pd(common->cm_id->verbs);
    if (!common->pd)
        return 1;

    /* Create a completion queue (CQ) where actual I/O 
        * completion metadata is placed. The metadata is packed into a structure 
        * called struct ibv_wc (wc = work completion). ibv_wc has detailed 
        * information about the work completion. An I/O request in RDMA world 
        * is called "work" ;)
        */
    common->send_cq = ibv_create_cq(common->cm_id->verbs, CQ_CAPACITY, NULL, NULL, 0); 
    if (!common->send_cq)
        return 1;
    common->recv_cq = ibv_create_cq(common->cm_id->verbs, CQ_CAPACITY, NULL, NULL, 0); 
    if (!common->recv_cq)
        return 1;

    if (config->shared_variables) {
        common->buf = config->shared_variables;
        common->shared_variables = true;
    } else {
        ret = allocate_variables(common, config);
        if (ret) {
            rdma_error("allocate_variables() failed");
            return 1;
        }
    }

    /* access permissions */
    int mr_flags = IBV_ACCESS_LOCAL_WRITE |
                   IBV_ACCESS_REMOTE_READ |
                   IBV_ACCESS_REMOTE_WRITE;

    /* register the memory buffer */
    common->mr = calloc(config->variables, sizeof(*common->mr));
    if (!common->mr)
        return 1;

    for (unsigned i = 0; i < config->variables; ++i) {
        local_var v = {i};
        uint32_t *addr = local_var_addr(common, v);

        common->mr[i] = ibv_reg_mr(common->pd, addr, sizeof (uint32_t), mr_flags);
        if (!common->mr[i]) {
            rdma_error("MR registration failed. errno=%d\n", errno);
            return 1;
        }
    }

    common->delay_sends_buf = malloc(DELAY_SEND_BUF_SIZE);
    if (!common->delay_sends_buf) {
        printf("Error allocating memory for delay sends.");
        return 1;
    }

    common->delay_sends_mr = ibv_reg_mr(common->pd, common->delay_sends_buf, DELAY_SEND_BUF_SIZE,
        IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE | IBV_ACCESS_REMOTE_READ);
    if (!common->delay_sends_mr) {
        printf("Error registering delay ops MR.\n");
        return 1;
    }

    common->inline_mr = ibv_reg_mr(common->pd, &common->local_buffer_for_inline_writes,
        sizeof (common->local_buffer_for_inline_writes),
        0);

    if (!common->inline_mr)
        return 1;

    /* Set up the queue pair (send, recv) queues and their capacity.
     * The capacity here is defined statically but this can be probed from the
     * device. We just use a small number  */
    qp_attr.cap.max_send_wr = 1024; /* Maximum send posting capacity */
    qp_attr.cap.max_send_sge = 2; /* Maximum SGE per send posting */
    qp_attr.cap.max_recv_wr = 1024; /* Maximum receive posting capacity */
    qp_attr.cap.max_recv_sge = 2; /* Maximum SGE per receive posting */
    qp_attr.cap.max_inline_data = sizeof(uint32_t);

    qp_attr.send_cq = common->send_cq; /* Where should I notify for send completion operations */
    qp_attr.recv_cq = common->recv_cq; /* Where should I notify for receive completion operations */
    qp_attr.qp_type = IBV_QPT_RC; /* QP type, RC = Reliable connection */

    int err = rdma_create_qp(common->cm_id, common->pd, &qp_attr); 
    if (err) {
        perror("rdma cm create qp error");
        return err;
    }

    assert(qp_attr.cap.max_inline_data >= sizeof(uint32_t));

    return 0;
}

struct pdata* prepare_common_metadata(struct common *common) {
    struct pdata *pdata = malloc(sizeof(struct pdata) + sizeof(struct remote_memory_metadata) * common->variables);

    pdata->variables = common->variables;

    pdata->delay_sends.addr = (uint64_t)(uintptr_t)common->delay_sends_buf;
    pdata->delay_sends.length = DELAY_SEND_BUF_SIZE;
    pdata->delay_sends.rkey = htonl(common->delay_sends_mr->rkey);

    for (int i = 0; i < common->variables; ++i) {
        const local_var v = {i};

        pdata->vars[i].addr = htole64((uintptr_t)local_var_addr(common, v));
        pdata->vars[i].length = htole32(common->mr[i]->length);
        /* Prepare remote key to provide to client */
        printf("rkey to provide for var %d: 0x%x\n", i, common->mr[i]->rkey);
        pdata->vars[i].rkey = htobe32(common->mr[i]->rkey); 
    }

    return pdata;
}

void print_gracefully_rdma_buffer_attr(struct common *common) {
    struct pdata *attr = common->pdata;

	if(!attr){
		rdma_error("Passed attr is NULL\n");
		return;
	}
	printf("---------------------------------------------------------\n");
    for (int i = 0; i < attr->variables; ++i) {
        printf("received remote buffer attr for var %d, addr: %p , len: %u , stag : 0x%x \n", 
                i,
                (void*) attr->vars[i].addr, 
                (unsigned int) attr->vars[i].length,
                attr->vars[i].rkey);
        assert(attr->vars[i].rkey);
    }
	printf("received remote delay-sends attr, addr: %p , len: %u , stag : 0x%x \n", 
			(void*) attr->delay_sends.addr, 
			(unsigned int) attr->delay_sends.length,
			attr->delay_sends.rkey);
    assert(attr->delay_sends.rkey);
	printf("---------------------------------------------------------\n");

}

void free_common(struct common *common) {
    int ret;

    for (int i = 0; i < common->variables; ++i) {
        ret = ibv_dereg_mr(common->mr[i]);
        assert(ret == 0);
    }

    free(common->mr);

    // Free allocated memory
    if (!common->shared_variables)
        // TODO free all allocations
        free(common->buf);

    ret = ibv_dereg_mr(common->delay_sends_mr);
    assert(ret == 0);
    free(common->delay_sends_buf);


    // Destroy the CM ID
    // Destroy the RDMA ID
    ret = rdma_destroy_id(common->cm_id);
    if (ret)  {
        perror("rdma_destroy_id");
        assert(0);
    }

    // Destroy the event channel
    rdma_destroy_event_channel(common->cm_channel);
}

/* The round trip time between the client and the server */
#define RTT_NS 3000

/* A CPU-bound delay using a busy-wait loop:
 * The delay is implemented using a busy-wait loop, where the CPU is continually 
 * polled until the desired delay duration is reached.
 */
void cpu_delay(uint64_t duration_ns)
{
    // No printing within this function as it changes the delay.
    // printf("\t\tDelay: %.3f milliseconds\n", double(duration_ns) / 1e6);
    struct timespec start, now;
    uint64_t diff_ns = 0;
    int ret = clock_gettime(CLOCK_MONOTONIC, &start);
    assert(!ret);

    while (diff_ns < duration_ns) {
        ret = clock_gettime(CLOCK_MONOTONIC, &now);
        assert(!ret);
        diff_ns = now.tv_sec * 1000000000 + now.tv_nsec - 
            (start.tv_sec * 1000000000 + start.tv_nsec);
    }
}

void random_cpu_delay()
{
    // Array of possible durations in nanoseconds
    const uint64_t durations_ns[] = {0, RTT_NS / 2, RTT_NS};
    int numDurations = sizeof(durations_ns) / sizeof(durations_ns[0]);

    uint64_t duration_ns = durations_ns[rand() % numDurations];
    
    if (!duration_ns)
        return;

    cpu_delay(duration_ns);
}

/* Add a random delay on the initiator side by posting large write operations
 * on the selected queue pair.
 */
void random_nic_delay(struct common *common) {
    struct ibv_send_wr send_wr = {}, *bad_send_wr;
    struct ibv_sge sge;

    const int num_operations = 1; // TODO: randomize
    int ret = 0;

	sge.addr = (uintptr_t)common->delay_sends_buf;
	sge.length = DELAY_SEND_BUF_SIZE - ((rand() % 4)  * 128);
	sge.lkey = common->delay_sends_mr->lkey;
	send_wr.wr_id = RANDOM_NIC_DELAY_WR_ID;
	send_wr.opcode = IBV_WR_RDMA_WRITE;
	send_wr.send_flags = 0; // no CQE unless there is an error
	send_wr.sg_list = &sge;
	send_wr.num_sge = 1;
	send_wr.wr.rdma.rkey = ntohl(common->pdata->delay_sends.rkey);
	send_wr.wr.rdma.remote_addr = common->pdata->delay_sends.addr;

	for (int i = 0; i < num_operations; ++i) {
		if ((ret = ibv_post_send(common->cm_id->qp, &send_wr, &bad_send_wr))) {
            printf("Error posting delay write operation.\n");
			exit(1);
		}
		send_wr.wr_id += 1;
	}
}

void random_delay_read(struct common *common) {
    struct ibv_send_wr send_wr = {}, *bad_send_wr;
    struct ibv_sge sge;

    const int num_operations = 1; // TODO: randomize
    int ret = 0;

	sge.addr = (uintptr_t)common->delay_sends_buf;
	sge.length = DELAY_SEND_BUF_SIZE - ((rand() % 4)  * 128);
	sge.lkey = common->delay_sends_mr->lkey;
	send_wr.wr_id = RANDOM_NIC_DELAY_WR_ID + 1;
	send_wr.opcode = IBV_WR_RDMA_READ;
	send_wr.send_flags = 0; // no CQE unless there is an error
	send_wr.sg_list = &sge;
	send_wr.num_sge = 1;
	send_wr.wr.rdma.rkey = ntohl(common->pdata->delay_sends.rkey);
	send_wr.wr.rdma.remote_addr = common->pdata->delay_sends.addr;

	for (int i = 0; i < num_operations; ++i) {
		if ((ret = ibv_post_send(common->cm_id->qp, &send_wr, &bad_send_wr))) {
            printf("Error posting delay write operation.\n");
			exit(1);
		}
		send_wr.wr_id += 1;
	}
}

#define DEFAULT_RDMA_PORT (20079)

struct common_setup_config default_config(uint32_t variables, const char *iteration) {
    struct common_setup_config config = {
        .variables = variables,
        .initiator_depth = 4,
        .responder_resources = 4,
        .iteration = iteration,
        .port = DEFAULT_RDMA_PORT
    };
    return config;
}