#pragma once

#include <assert.h>
#include <endian.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>

#include <infiniband/verbs.h>


#define rdma_error(msg, args...) \
    do { \
        fprintf(stderr, "%s : %d : ERROR : "msg, __FILE__, __LINE__, ## args); \
    } while (0)

#ifdef ACN_RDMA_DEBUG 
#define debug(msg, args...) \
    do { \
        printf("DEBUG: "msg, ## args); \
    } while (0)
#else
#define debug(msg, args...)
#endif

struct remote_memory_metadata {
    uint64_t addr;
    uint32_t length;
    uint32_t rkey;
};

#define MAX_VARIABLES 8

struct pdata {
    /* Number of variables */
    uint8_t variables;
    uint8_t reserved[3]; // Padding

    struct remote_memory_metadata delay_sends;

    /* variables entries of remote_memory_metadata */
    struct remote_memory_metadata vars[0];
};

struct common_setup_config {
    uint32_t variables;
    uint32_t initiator_depth;
    uint32_t responder_resources;
    const char *iteration;
    bool randomize_variable_home;
    uint16_t port;
    uint32_t **shared_variables;
    uint32_t variables_initialization_value;
};

/* Common structure to both client and server */
struct common {
   	struct rdma_event_channel		*cm_channel;
   	struct rdma_cm_id				*cm_id;
   	struct ibv_pd					*pd;
   	struct ibv_cq					*send_cq, *recv_cq; 
   	struct pdata					*pdata;
    uint8_t variables;
    bool shared_variables;
    uint32_t **buf;
    struct ibv_mr **mr; /* MR for every variable */
    uint8_t *delay_sends_buf;
    struct ibv_mr *delay_sends_mr;
    uint32_t local_buffer_for_inline_writes;
    struct ibv_mr *inline_mr;
};

int init_common_channel(struct common *common);
int init_common(struct common *common, struct common_setup_config* config);
struct pdata *prepare_common_metadata(struct common *common);
void print_gracefully_rdma_buffer_attr(struct common *common);
void free_common(struct common *common);

void cpu_delay(uint64_t duration_ns);
void random_cpu_delay();
void random_nic_delay(struct common *common);
void random_delay_read(struct common *common);

#define DELAY_SEND_BUF_SIZE (4 * 65536)
#define RANDOM_NIC_DELAY_WR_ID 0x10

#define DISCONNECT_IMMEDIATE_VALUE 0x4947414d

#define CQ_CAPACITY (1600)

struct common_setup_config default_config(uint32_t variables, const char *iteration);

typedef struct {
    uint8_t id;
} local_var;

typedef struct {
    uint8_t id;
} remote_var;

static inline uint32_t *local_var_addr(struct common *common, local_var local) {
	assert(local.id < common->variables);

    return common->buf[local.id];
}

static inline void set_remote_var(struct common *common, struct ibv_send_wr *wr, remote_var remote) {
	assert(remote.id < common->pdata->variables);

    wr->wr.rdma.rkey = be32toh(common->pdata->vars[remote.id].rkey);
    wr->wr.rdma.remote_addr = le64toh(common->pdata->vars[remote.id].addr);
}

static inline void set_local_var(struct common *common, struct ibv_sge *sge, local_var local) {
    sge->addr   = (uint64_t)(uintptr_t)local_var_addr(common, local);
	sge->length = sizeof(uint32_t);
	sge->lkey   = common->mr[local.id]->lkey; /* Key of the local Memory Region */
}

static inline uint32_t load_local(struct common *common, local_var local) {
    return *local_var_addr(common, local);
}

static inline void store_local(struct common *common, local_var local, uint32_t value) {
    *local_var_addr(common, local) = value;
}