/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make crma1_client
 *** gcc -g -o crma1_client crma1_client.c rdma_setup_client.c handle_work_completion.c disconnect_cleanup.c -lrdmacm -libverbs
 *
 * usage：
 * ./crma1_client 192.168.0.2 192.168.0.3 <index> // to sgx02 & sgx03
 * 
 *      P1       │ P2       │ P3
 *      init x=0 │ init y=0 │ init z=0
 *     ──────────┤──────────┤─────────────
 *     x := y²   │ y  := 2  │  z := 1
 *     flush(2)  │          |  b := z
 *     z³ := x   │          │
 *     a := x    │          │
 *     ──────────┤──────────┤──────────
 *     (a,b)=(2,0) not possible
 *
 *    ╔════════════════╗              ╔═════════════════╗
 *    ║ Client (local) ║              ║ Server1 (remote)║
 *    ╚════════════════╝              ╚═════════════════╝
 *    │ x ...    buf[0]│<── x:=y² ────│ y ... buf[0]    │
 *    │                │              └─────────────────┘
 *    │                │              ╔═════════════════╗
 *    │                │              ║ Server2 (remote)║
 *    │                │              ╚═════════════════╝
 *    │                │ ── z³:=x ────│ z ... buf[0]    │
 *    └────────────────┘              └─────────────────┘
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"

#include <assert.h>

int main(int argc, char *argv[]) {

	char *itr = (argc > 3) ? argv[3] : "(null)";	
    // Seed the random number generator
    srand(time(0));
	struct client_common client1 = {}, client2 = {};

    if (argc < 4) {
        fprintf(stderr, "Usage:./crma1_client <server1 ip> <server2 ip> [iteration]\n");
        exit(1);
    }
/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */

    struct common_setup_config config = default_config(3 /* variables */, itr);
    rdma_setup_client(&client1, argv[1], config);

    config.shared_variables = client1.common.buf;
    config.port++;
    rdma_setup_client(&client2, argv[2], config);
   
    // Define pointer aliases
    const local_var x = {0}, /* sync1 = {1}, */ sync2 = {2};
    const remote_var server1_y = {0}, server2_z = {0};

// Begin of Litmus test
    // write_inline_w_imm(&client.common, server1_sync, 0, 0x80000000ul, true /* fence */);
    // ret = handle_work_completion_status_events(client.common.send_cq, 1);
    // assert(ret == 1);

    /*  x := y² */
    read_to_from(&client1.common, x, server1_y);

    /* Wait for the client to indicate the completion of the operation  */
    int ret = handle_work_completion_status_events(client1.common.send_cq, 1);
    if (ret != 1) {
        rdma_error("Failed to get 1 work completion from client, ret = %d \n", ret);
        return ret;
    }

    /* z³ := x */
    write_to_from(&client2.common, server2_z, x, false);

    uint32_t a = load_local(&client1.common, x);

// End of Litmus test

    /* Wait for the sync messages */
	ret = handle_work_completion_status_events(client1.common.recv_cq, 1);
    assert(ret == 1);
    ret = handle_work_completion_status_events(client2.common.recv_cq, 1);
    assert(ret == 1);

    uint32_t b = load_local(&client2.common, sync2);
  
    // Call the disconnect() function
    int result = disconnect(&client1);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }
    result = disconnect(&client2);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

    printf("[%s] Upon Litmus Test execution: (a,b) == (%u,%u) [x == %u]\n",
        itr, a, b, load_local(&client1.common, x));

    // Call the cleanup function
    result = cleanup(&client1);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }
    result = cleanup(&client2);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }
    return 0;
}



	
	

