/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make crma2_client
 *** gcc -g -o crma2_client crma2_client.c rdma_setup_client.c handle_work_completion.c disconnect_cleanup.c -lrdmacm -libverbs
 *
 * usage：
 * ./crma2_client 192.168.0.2 <index> // to sgx02
 * 
 *
 ◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤
 * The server-side metadata (buffer/rkey) is transmitted to the client upon accepting 
 * the connection (rdma_accept), using the rep_pdata mechanism.
 *
 * On the client side, the buffer information metadata is communicated to the server 
 * through a send/recv exchange.
 ◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣
 *
 *     P1         │   P2
 *     init x,y=0 │  init z,w=0
 *     ───────────┤─────────────
 *     z² := x    │  w := 1 
 *     w² := y    │  a := z
 *     y  := 2    │  b := w 
 *     x  := 2    │ 
 *     ─────────────────────────
 *       (a,b)=(2,0) not permitted
 *
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client [local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ x ... buf[0]   │─── z²:= x ──>│ z ... buf[0]   │
 *    ├────────────────┤              ├────────────────┤
 *    │ y ... buf[1]   │─── w²:= y ──>│ w ... buf[1]   │ 
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"

#include <assert.h>

int main(int argc, char *argv[]) {

	char *itr = (argc > 2) ? argv[2] : "(null)";	
    // Seed the random number generator
    srand(time(0));
	struct client_common client = {};

	bool fence = false;


    	
/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */
    rdma_setup_client(&client, argv[1], default_config(3 /* variables */, itr));

    const local_var  x    = {0},
                     y    = {1},
                     sync = {2};
                    
    const remote_var z    = {0},
                     w    = {1};

    (void)sync;

    // random_cpu_delay();
    random_nic_delay(&client.common);
    
// Begin of Litmus test
    
	/* z^2 := x */
    write_to_from(&client.common, z, x, fence);


	/* w^2 := y */
    write_to_from(&client.common, w, y, fence);
    
    
    store_local(&client.common, y, 2);
    
    store_local(&client.common, x, 2);


	/* Wait for the server to indicate the completion of the operation  */
	int ret = handle_work_completion_status_events(client.common.send_cq, 2);
	if (ret != 2) {
		rdma_error("Failed to get 2 work completions from server, ret = %d \n", ret);
		return ret;
	}
    
// End of Litmus test

    /* Wait for the sync message */
    int ret2 = handle_work_completion_status_events(client.common.recv_cq, 1);
    assert(ret2 == 1);

    // Call the disconnect() function
    int result = disconnect(&client);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }



    // Call the cleanup function
    result = cleanup(&client);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }    

    return 0;
}



	
	

