/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make fig2e_client
 * 
 * usage：
 * ./fig2e_client 192.168.0.2 192.168.0.3 <index> // to sgx02 & sgx03
 *
 *     P1         │   P2        │   P3
 *     init x=0   │  init z=1   │  init y=2 
 *     ───────────┤─────────────┤───────────
 *     x  := z²   │             │
 *     x  := y³   │             │
 *     poll(2)    │             │
 *     poll(3)    │             │
 *     ──────────────────────────────────────
 *       x == {1,2} permitted 
 *
 *
 *
 *    ╔════════════════╗              ╔═════════════════╗              ╔═════════════════╗
 *    ║ Client [local) ║              ║ Server1 (remote)║              ║ Server2 (remote)║
 *    ╚════════════════╝              ╚═════════════════╝              ╚═════════════════╝
 *    │ x ... buf[0]   │<── x:=z² ────│ z ... buf[0]    │  ┌ x:= y³ ───│ y ... buf[0]    │
 *    └────────────────┘              └─────────────────┘  │           └─────────────────┘ 
 *            └──<─────────────────────────────────────────┘
 *
 * 
 * 
 *
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"

#include <assert.h>

int main(int argc, char *argv[]) {
	char *itr = (argc > 3) ? argv[3] : "(null)";	

    // Seed the random number generator
    srand(time(0));
    struct client_common client1 = {}, client2 = {};

    if (argc < 4) {
        fprintf(stderr, "Usage:./fig2e_client <server1 ip> <server2 ip> [iteration]\n");
        exit(1);
    }

/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */

    struct common_setup_config config = default_config(1 /* variables */, itr);
    rdma_setup_client(&client1, argv[1], config);

    config.shared_variables = client1.common.buf;
    config.port++;
    rdma_setup_client(&client2, argv[2], config);
   
    // Define pointer aliases
    const local_var x = {0};
    const remote_var server1_z = {0}, server2_y = {0};

/*
 ◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤
 ◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣
	
  _     _ _                         _____         _   
 | |   (_) |_ _ __ ___  _   _ ___  |_   _|__  ___| |_ 
 | |   | | __| '_ ` _ \| | | / __|   | |/ _ \/ __| __|
 | |___| | |_| | | | | | |_| \__ \   | |  __/\__ \ |_ 
 |_____|_|\__|_| |_| |_|\__,_|___/   |_|\___||___/\__|

*/

	printf("\t\t============fig2e: client side ===================\n");

    if (rand() % 2 == 0)
        random_nic_delay(&client1.common);

    /* x  := z² */
    read_to_from(&client1.common, x, server1_z);

    /* x  := y³ */
    read_to_from(&client2.common, x, server2_y);

	int ret = handle_work_completion_status_events(client1.common.send_cq, 1);
    if (ret != 1) {
        rdma_error("Failed to get 1 work completion from client 1, ret = %d \n", ret);
        return ret;
    }

    ret = handle_work_completion_status_events(client2.common.send_cq, 1);
    if (ret != 1) {
        rdma_error("Failed to get 1 work completion from client 2, ret = %d \n", ret);
        return ret;
    }

/*
 ◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤
 ◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣
*/    

// Call the disconnect() function
    int result = disconnect(&client1);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }
    result = disconnect(&client2);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

    printf("[%s] Upon Litmus Test execution: x == %u\n",
        itr, load_local(&client1.common, x));

    // Call the cleanup function
    result = cleanup(&client1);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }
    result = cleanup(&client2);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }
    return 0;
}







