/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make fig3e_client
 *** gcc -g -o fig3e_client fig3e_client.c rdma_setup_client.c handle_work_completion.c disconnect_cleanup.c -lrdmacm -libverbs
 *
 * usage：
 * ./fig3e_client 192.168.0.2 <index> // to sgx02
 * 
 *
 ◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤◢◤
 * The server-side metadata (buffer/rkey) is transmitted to the client upon accepting 
 * the connection (rdma_accept), using the rep_pdata mechanism.
 *
 * On the client side, the buffer information metadata is communicated to the server 
 * through a send/recv exchange.
 ◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣◥◣
 *
 *     P1         │   P2
 *     init y=w=0 │  init x=z=0
 *     ───────────┤─────────────
 *     x² := 1    │  y¹ := 1
 *     c  := z²   │  d  := w¹
 *     poll(2)    │  poll(1) 
 *     poll(2)    │  poll(1) 
 *     a  := y    │  b := x
 *     ─────────────────────────
 *            a=b=0 not possible
 *
 *
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client [local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ y ... buf[0]   │─── x²:=1 ───>│ x ... buf[0]   │
 *    ├────────────────┤              ├────────────────┤ 
 *    │ w ... buf[1]   │─── d:=w¹ ───>│ d ... buf[1]   │
 *    ├────────────────┤              ├────────────────┤ 
 *    │ c ... buf[2]   │<── c:=z² ────│ z ... buf[2]   │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"
#include <assert.h>

int main(int argc, char *argv[]) {

	char *itr = (argc > 2) ? argv[2] : "(null)";	
    // Seed the random number generator
    srand(time(0));
	struct client_common client = {};

	bool fence = false;


    	
/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */
    rdma_setup_client(&client, argv[1], default_config(4 /* variables */, itr));
   
    // Define pointer aliases
    const local_var y    = {0},
                    w    = {1},
                    c    = {2},
                    sync = {3};
    const remote_var x = {0},
                     d = {1},
                     z = {2};

    (void)d;
    (void)w;

// Begin of Litmus test
    
	/* x^2 := 1 */
    write_inline(&client.common, x, 1, fence);

	/* c  := z²  */
    read_to_from(&client.common, c, z);

	/* Wait for the server to indicate the completion of the operations  */
	int ret = handle_work_completion_status_events(client.common.send_cq, 2);
	if (ret != 2) {
		rdma_error("Failed to get 2 work completions from server, ret = %d \n", ret);
		return ret;
	}

    uint32_t a = load_local(&client.common, y);

    
// End of Litmus test

    /* Wait for server to complete before disconnecting. */
	ret = handle_work_completion_status_events(client.common.recv_cq, 1);
    assert(ret == 1);

    // Call the disconnect() function
    int result = disconnect(&client);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

    uint32_t b = load_local(&client.common, sync);

	printf("Upon Litmus Test execution: (a,b) == (%u,%u)\n", a, b);

    // Call the cleanup function
    result = cleanup(&client);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }    

    return 0;
}



	
	

