/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make gfg1_client
 *** gcc -g -o gfg1_client gfg1_client.c rdma_setup_client.c handle_work_completion.c disconnect_cleanup.c -lrdmacm -libverbs
 *
 * usage：
 * ./gfg1_client 192.168.0.2 <index> // to sgx02
 * 
 *     P1        │   P2
 *     init x=-1 │  init w,z=0
 *          y= 0 |
 *     ──────────┤─────────────
 *     x  := z²  │  w  := 1
 *     rfence    │  w  := 2
 *     y  := w²  │  x¹ := 3
 *     a  := x   │
 *     b  := x   |
 *     c  := y   |
 *     ─────────────────────────
 *     (a,b,c)=(3,0,1) possible
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client (local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ x ... buf[0]   │<── x:=z² ────│ z ... buf[0]   │
 *    │                │<── x²:=3 ──  │                │
 *    ├────────────────┤              ├────────────────┤ 
 *    │ y ... buf[1]   │<── y:=w² ────│ w ... buf[1]   │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"

#include <assert.h>

int main(int argc, char *argv[]) {

	char *itr = (argc > 2) ? argv[2] : "(null)";	
    // Seed the random number generator
    srand(time(0));
	struct client_common client = {};
    struct ibv_wc wc[2];
    int received_wcs = 0;
    int ret;

/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */

    struct common_setup_config config = default_config(3 /* variables */, itr);
    // config.initiator_depth = 1;
    rdma_setup_client(&client, argv[1], config);
   
    // Define pointer aliases
    const local_var x = {0}, y = {1};
                    // sync = {2};
    const remote_var z = {0},
                     w = {1};

// Begin of Litmus test
    store_local(&client.common, x, -1);

    int32_t a = -1, b = -1, c = -1;

    /* x  := z²  */
    read_to_from(&client.common, x, z);

    random_delay_read(&client.common);

    /* rfence; y := w² */
    read_to_from_fence(&client.common, y, w, true);

    /* a := x  */
    do {
        a = load_local(&client.common, x);
    } while (a == -1);

    do {
        ret = handle_work_completion_status_events_wc(client.common.send_cq, 0, 2 - received_wcs, &wc[received_wcs]);
        if (ret < 0) {
            rdma_error("Failure in handle work completion ret = %d.", ret);
            return ret;
        } else if (ret > 0) {
            assert(received_wcs + ret <= 2);
            for (int i = received_wcs; i < received_wcs + ret; ++i)
                assert(wc[i].opcode == IBV_WC_RDMA_READ);
            received_wcs += ret;
        }

        b = load_local(&client.common, x);
    } while (b == a && received_wcs < 2);

    /* Wait for the client to indicate the completion of the operation  */
    ret = handle_work_completion_status_events(client.common.send_cq, 2 - received_wcs);
    if (ret != 2 - received_wcs) {
        rdma_error("Failed to get 1 work completion from client, ret = %d \n", ret);
        return ret;
    }

    c = load_local(&client.common, y);

// End of Litmus test

    /* Wait for the sync message */
	ret = handle_work_completion_status_events(client.common.recv_cq, 1);
    assert(ret == 1);
  
    // Call the disconnect() function
    int result = disconnect(&client);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

	printf("Upon Litmus Test execution: (a,b,c) == (%u,%u,%u) [x == %u, y == %u]\n",
        a, b, c, load_local(&client.common, x), load_local(&client.common, y));

    // Call the cleanup function
    result = cleanup(&client);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }    

    return 0;
}



	
	

