/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make gfp2_client
 *** gcc -g -o gfp2_client gfp2_client.c rdma_setup_client.c handle_work_completion.c disconnect_cleanup.c -lrdmacm -libverbs
 *
 * usage：
 * ./gfp2_client 192.168.0.2 <index> // to sgx02
 * 
 *     P1        │   P2
 *     init x=-1 │  init y,z=0
 *     ──────────┤─────────────
 *     x  := z²  │  a  := y
 *     rfence    │  x¹ := 2
 *     y² := 1   │
 *     b  := x   │
 *     c  := x   |
 *     ─────────────────────────
 *     (a,b,c)=(1,2,0) possible
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client [local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ x ... buf[0]   │<── x:=z² ────│ z ... buf[0]   │
 *    │                │<── x²:=2 ────│                │
 *    │                │              ├────────────────┤ 
 *    │                │─── y:=1 ────>│ y ... buf[1]   │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"

#include <assert.h>

int main(int argc, char *argv[]) {

	char *itr = (argc > 2) ? argv[2] : "(null)";	
    // Seed the random number generator
    srand(time(0));
	struct client_common client = {};
    struct ibv_wc wc;
    int received_wcs = 0;
    int ret;

/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */

    struct common_setup_config config = default_config(2 /* variables */, itr);
    // config.initiator_depth = 1;
    rdma_setup_client(&client, argv[1], config);
   
    // Define pointer aliases
    const local_var x = {0},
                    sync = {1};
    const remote_var z = {0},
                     y = {1};

// Begin of Litmus test
    store_local(&client.common, x, -1);

    random_delay_read(&client.common);

	/* x  := z²  */
    read_to_from(&client.common, x, z);

    /* y² := 1  */
    /* Can't batch with previous op because then BlueFlame isn't used. */
    write_inline(&client.common, y, 1, true /* rfence */);


    /* b  := x  */
    int32_t b;
    do {
        b = load_local(&client.common, x);
    } while (b == -1);

    int32_t c = -1;
    do {
        ret = handle_work_completion_status_events_wc(client.common.send_cq, 0, 1 - received_wcs, &wc);
        if (ret < 0) {
            rdma_error("Failure in handle work completion ret = %d.", ret);
            return ret;
        } else if (ret > 0) {
            received_wcs += ret;
            assert(wc.opcode == IBV_WC_RDMA_READ);
        }

        c = load_local(&client.common, x);
    } while (c == b && received_wcs == 0);

    // printf("Exit loop with received_wcs == %d\n", received_wcs);

	/* Wait for the client to indicate the completion of the operation  */
	ret = handle_work_completion_status_events(client.common.send_cq, 2 - received_wcs);
	if (ret != 2 - received_wcs) {
		rdma_error("Failed to get 1 work completion from client, ret = %d \n", ret);
		return ret;
	}

// End of Litmus test

    /* Wait for the sync message */
	ret = handle_work_completion_status_events(client.common.recv_cq, 1);
    assert(ret == 1);
  
    uint32_t a = load_local(&client.common, sync);

    // Call the disconnect() function
    int result = disconnect(&client);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

	printf("Upon Litmus Test execution: (a,b,c) == (%u,%u,%u) [x == %u]\n",
        a, b, c, load_local(&client.common, x));

    // Call the cleanup function
    result = cleanup(&client);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }    

    return 0;
}



	
	

