/* 
 *  █▀ █▀▀ █▀█ █░█ █▀▀ █▀█ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  ▄█ ██▄ █▀▄ ▀▄▀ ██▄ █▀▄ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build:
 *   make iriw1_server
 * 
 * usage: 
 * ./iriw1_server <index>
 *
 *     P1                   │  P2
 *     init x=0             │  init y=0
 *     ─────────┤───────────|───────────┤──────────
 *     x  := 1  |  a := y²  │  c := x¹  | y  := 1
 *              │  b := x   |  d := y   |
 *     ──────────────────────────────────────────────
 *            (a,b,c,d) == (1,0,1,0) possible
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client [local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ a ... buf[0]   │<─── a:=y² ───│ y ... buf[0]   │
 *    └────────────────┘              └────────────────┘
 *    │ x ... buf[1]   │──── c:=x¹ ──>│ c ... buf[1]   │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_server.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_server.h"
#include "rdma_setup_server.h"
#include "write.h"
#include "read.h"

#include <assert.h>

void *secondary_thread_func(void *arg) {
    struct server_common *server = (struct server_common *)arg;

    const local_var y = {0};

    store_local(&server->common, y, 1);

    return NULL;
}

int main(int argc, char *argv[]) 
{ 
	char *iter = (argc > 1) ? argv[1] : "(null)";
	
    // Seed the random number generator
    srand(time(0));

	struct server_common server = {};

/*
 - Setup RDMA resources
 - Exchange information about the local buffer between the client and server 
 */
	rdma_setup_server(&server, default_config(3 /* variables */, iter));

    // Start helper thread
    pthread_t secondary_thread;
    int ret = pthread_create(&secondary_thread, NULL, secondary_thread_func, &server);
    assert(ret == 0);

    // Define pointer aliases
    const local_var y = {0},
                    c = {1};
    const remote_var a = {0},
                     x = {1};

    (void)a;

    // Local
    uint32_t d = -1;
	
/*
 ◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤
 ◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣
	
  _     _ _                         _____         _   
 | |   (_) |_ _ __ ___  _   _ ___  |_   _|__  ___| |_ 
 | |   | | __| '_ ` _ \| | | / __|   | |/ _ \/ __| __|
 | |___| | |_| | | | | | |_| \__ \   | |  __/\__ \ |_ 
 |_____|_|\__|_| |_| |_|\__,_|___/   |_|\___||___/\__|

*/

    while (load_local(&server.common, c) == 0) {
        /* c  := x¹  */
        read_to_from(&server.common, c, x);

        d = load_local(&server.common, y);

        /* Wait for the server to indicate the completion of the operations  */
        ret = handle_work_completion_status_events(server.common.send_cq, 1);
        if (ret != 1) {
            rdma_error("Failed to get 2 work completions from server, ret = %d \n", ret);
            return ret;
        }
    }

/*
 ◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤
 ◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣
*/ 

    ret = pthread_join(secondary_thread, NULL);
    assert(ret == 0);

    // Call the disconnect() function
    int result = disconnect(&server);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

	printf("Upon Litmus Test execution: (c,d) == (%u,%u)\n",
        load_local(&server.common, c), d);


    // Call the cleanup function
    result = cleanup(&server);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }

		
    return 0;
}
