/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make iriw2_client
 *
 * usage：
 * ./iriw2_client 192.168.0.2 <index> // to sgx02
 * 
 *     P1                   │  P2
 *     init x=0             │  init y=0
 *     ─────────┤───────────|───────────┤──────────
 *     x  := 1  |  a := y²  │  c := x¹  | y  := 1
 *              │  poll(2)  |  poll(1)  |
 *              │  b := x   |  d := y   |
 *     ────────────────────────────────────────────
 *            (a,b,c,d) == (1,0,1,0) not possible
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client [local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ a ... buf[0]   │<─── a:=y² ───│ y ... buf[0]   │
 *    └────────────────┘              └────────────────┘
 *    │ x ... buf[1]   │──── c:=x¹ ──>│ c ... buf[1]   │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"

#include <assert.h>

// Helper pthread
void *secondary_thread_func(void *arg) {
    struct client_common *client = (struct client_common *)arg;

    const local_var x = {1};

    store_local(&client->common, x, 1);

    return NULL;
}

int main(int argc, char *argv[]) {

	char *itr = (argc > 2) ? argv[2] : "(null)";	
    // Seed the random number generator
    srand(time(0));
	struct client_common client = {};

/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */
    rdma_setup_client(&client, argv[1], default_config(2 /* variables */, itr));

    // Start helper thread
    pthread_t secondary_thread;
    int ret = pthread_create(&secondary_thread, NULL, secondary_thread_func, &client);
    assert(ret == 0);

    // Define pointer aliases
    const local_var a = {0},
                    x = {1};
    const remote_var y = {0},
                     c = {1};

    (void)c;

    // Local
    uint32_t b;

// Begin of Litmus test


    while (load_local(&client.common, a) == 0) {
        /* a  := y²  */
        read_to_from(&client.common, a, y);

        /* Wait for the server to indicate the completion of the operations  */
        ret = handle_work_completion_status_events(client.common.send_cq, 1);
        if (ret != 1) {
            rdma_error("Failed to get 2 work completions from server, ret = %d \n", ret);
            return ret;
        }

        b = load_local(&client.common, x);
    }

// End of Litmus test

    ret = pthread_join(secondary_thread, NULL);
    assert(ret == 0);

    // Call the disconnect() function
    ret = disconnect(&client);
    if (ret != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", ret);
    }


	printf("Upon Litmus Test execution: (a,b) == (%u,%u)\n",
        load_local(&client.common, a), b);

    // Call the cleanup function
    ret = cleanup(&client);
    if (ret != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", ret);
    }    

    return 0;
}



	
	

