/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make qp1_client
 *** gcc -g -o qp1_client qp1_client.c rdma_setup_client.c handle_work_completion.c disconnect_cleanup.c -lrdmacm -libverbs
 *
 * usage：
 * ./qp1_client 192.168.0.2 <index> // to sgx02
 * 
 *     P1                │   P2
 *     init y=0          │  init x,z=0
 *     ──────────────────┤─────────────
 *     x² := 1 │ a  := y │  b  := z
 *     poll(2) │ z² := 1 │  c  := x
 *     y := 1  │         │
 *     ────────────────────────────────
 *     (a,b,c)=(1,1,0) possible
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client (local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ sync ... buf[0]│ ── x²:=1 ───>│ x ... buf[0]   │
 *    │                │              ├────────────────┤ 
 *    │                │ ── z²:=1 ────│ z ... buf[1]   │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"

#include <assert.h>

struct shared_memory {
    bool c2_ready;
    uint32_t y;
};

struct client2_args {
    struct shared_memory shm;
    const char *server_address;
};

void *secondary_thread_func(void *arg) {
    struct client_common client = {};
    struct client2_args *c2 = (struct client2_args *)arg;

    struct common_setup_config config = default_config(1 /* variables */, "client2");
    config.port++;
    rdma_setup_client(&client, c2->server_address, config);
    c2->shm.c2_ready = true;

    const remote_var z = {0};

    uint32_t a = 0;

    do {
        a = c2->shm.y;
        random_cpu_delay();
    } while (a == 0);

    write_inline(&client.common, z, 1, false);    

    /* Wait for the client to indicate the completion of the operation  */
    int ret = handle_work_completion_status_events(client.common.send_cq, 1);
    if (ret != 1) {
        rdma_error("Failed to get 1 work completion from client, ret = %d \n", ret);
        exit(ret);
    }

// End of Litmus test

    /* Wait for the sync message */
	ret = handle_work_completion_status_events(client.common.recv_cq, 1);
    assert(ret == 1);
  
    // Call the disconnect() function
    int result = disconnect(&client);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

	printf("Upon Litmus Test second thread execution: a == %u\n", a);

    // Call the cleanup function
    result = cleanup(&client);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }

    return NULL;
}

int main(int argc, char *argv[]) {

	char *itr = (argc > 2) ? argv[2] : "(null)";	
    // Seed the random number generator
    srand(time(0));
	struct client_common client = {};

    struct client2_args c2 = {};
    c2.server_address = argv[1];

    pthread_t secondary_thread;
    int ret = pthread_create(&secondary_thread, NULL, secondary_thread_func, &c2);
    assert(ret == 0);

/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */

    struct common_setup_config config = default_config(1 /* variables */, itr);
    rdma_setup_client(&client, argv[1], config);
   
    // Define pointer aliases
    const remote_var x = {0}, server1_sync = {1};

// Begin of Litmus test
    while (!c2.shm.c2_ready)
        random_cpu_delay();

    write_inline_w_imm(&client.common, server1_sync, 0, 0x80000000ul, true /* fence */);
    ret = handle_work_completion_status_events(client.common.send_cq, 1);
    assert(ret == 1);

    /* x² := 1 */
    write_inline(&client.common, x, 1, false);

    /* Wait for the client to indicate the completion of the operation  */
    ret = handle_work_completion_status_events(client.common.send_cq, 1);
    if (ret != 1) {
        rdma_error("Failed to get 1 work completion from client, ret = %d \n", ret);
        return ret;
    }

    /* y := 1 */
    c2.shm.y = 1;

// End of Litmus test

    /* Wait for the sync message */
	ret = handle_work_completion_status_events(client.common.recv_cq, 1);
    assert(ret == 1);
  
    // Call the disconnect() function
    int result = disconnect(&client);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

    ret = pthread_join(secondary_thread, NULL);
    assert(ret == 0);

    // Call the cleanup function
    result = cleanup(&client);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }    

    return 0;
}



	
	

