/* 
 *  █▀ █▀▀ █▀█ █░█ █▀▀ █▀█ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  ▄█ ██▄ █▀▄ ▀▄▀ ██▄ █▀▄ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build:
 * gcc -o qp1_server qp1_server.c -lrdmacm -libverbs
 * 
 * usage: 
 * ./qp1_server <index>
 * 
 *     P1                │   P2
 *     init y=0          │  init x,z=0
 *     ──────────────────┤─────────────
 *     x² := 1 │ a  := y │  b  := z
 *     poll(2) │ z² := 1 │  c  := x
 *     y := 1  │         │
 *     ────────────────────────────────
 *     (a,b,c)=(1,1,0) possible
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client (local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ sync ... buf[0]│ ── x²:=1 ───>│ x ... buf[0]   │
 *    │                │              ├────────────────┤ 
 *    │                │ ── z²:=1 ────│ z ... buf[1]   │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_server.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_server.h"
#include "rdma_setup_server.h"
#include "read.h"
#include "write.h"

void *setup_second_server(void *arg) {
    struct server_common *server = (struct server_common *)arg;

    struct common_setup_config config = default_config(1 /* variables */, "server2");
    config.port++;
    rdma_setup_server(server, config);

    return NULL;
}

int main(int argc, char *argv[]) 
{ 
	char *iter = (argc > 1) ? argv[1] : "(null)";
	
    // Seed the random number generator
    srand(time(0));

	struct server_common server1 = {},
                         server2 = {};

    pthread_t secondary_thread;
    int ret = pthread_create(&secondary_thread, NULL, setup_second_server, &server2);
    assert(ret == 0);

/*
 - Setup RDMA resources
 - Exchange information about the local buffer between the client and server 
 */
	rdma_setup_server(&server1, default_config(2 /* variables */, iter));

    ret = pthread_join(secondary_thread, NULL);
    assert(ret == 0);

    // Define pointer aliases
    const local_var server1_x = {0},
                    server2_z = {0};
    const remote_var sync = {0};

    /* Wait for the sync message */
	ret = handle_work_completion_status_events(server1.common.recv_cq, 1);
    assert(ret == 1);

/*
 ◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤
 ◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣
	
  _     _ _                         _____         _   
 | |   (_) |_ _ __ ___  _   _ ___  |_   _|__  ___| |_ 
 | |   | | __| '_ ` _ \| | | / __|   | |/ _ \/ __| __|
 | |___| | |_| | | | | | |_| \__ \   | |  __/\__ \ |_ 
 |_____|_|\__|_| |_| |_|\__,_|___/   |_|\___||___/\__|

*/
    random_cpu_delay();

	/* b := z  */
    uint32_t b = load_local(&server2.common, server2_z);

    random_cpu_delay();

    /* c := x  */
    uint32_t c = load_local(&server1.common, server1_x);

/*
 ◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤
 ◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣
*/ 
    write_inline_w_imm(&server1.common, sync, b, 0x80000000ul, true /* fence */);
    write_inline_w_imm(&server2.common, sync, c, 0x80000000ul, true /* fence */);

    // Call the disconnect() function
    int result = disconnect(&server1);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect1 failed with error code %d\n", result);
    }

    result = disconnect(&server2);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect2 failed with error code %d\n", result);
    }

	printf("[%s] Upon Litmus Test execution: (b,c) == (%u, %u) [x == %u, z == %u]\n",
        iter, b, c, load_local(&server1.common, server1_x), load_local(&server2.common, server2_z));


    // Call the cleanup function
    result = cleanup(&server1);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }
    result = cleanup(&server2);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }
		
    return 0;
}
