// rdma_setup_client.c

#include "rdma_common_client.h"
#include "handle_work_completion.h"

#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <rdma/rdma_cma.h>

void rdma_setup_client(struct client_common *client, const char *server_address,
                       struct common_setup_config config) {
    struct rdma_conn_param conn_param = {};
    struct addrinfo hints = {
        .ai_family = AF_INET,
        .ai_socktype = SOCK_STREAM
    };
    struct addrinfo					*res, *t;
    struct ibv_send_wr send_wr;
    struct ibv_send_wr *bad_send_wr;
    struct rdma_cm_event *event;
    struct ibv_recv_wr recv_wr = {}, *bad_recv_wr;
    struct ibv_sge sge;

	printf("\n\n==================== client side [%s]===================\n", config.iteration);

    /* Set up RDMA CM structures */
    if (init_common_channel(&client->common)) {
        rdma_error("Failed to initialize common channel\n");
        exit(1);
    }

    /* cm_id is the connection identifier (like socket) which is used
     * to define an RDMA connection.
     */
    int err = rdma_create_id(client->common.cm_channel, &client->common.cm_id, NULL, RDMA_PS_TCP);
    if (err) {
        rdma_error("rdma_create_id() failed with error %d\n", err);
        exit(err);
    }

    /* Note: port 20079 doesn't equal to the socket port in TCP/IP,
     * in RoCEv2, all of the packets use port 4791,
     * port 20079 here indicates a higher level abstraction port
     */

    char port_string[16] = "";
    snprintf(port_string, sizeof(port_string), "%u", config.port);

    int n = getaddrinfo(server_address, port_string, &hints, &res);
    if (n < 0) {
        rdma_error("getaddrinfo() failed with error %d\n", n);
        exit(1);
    }

    /* Resolve server address and route */
    for (t = res; t; t = t->ai_next) {
        err = rdma_resolve_addr(client->common.cm_id, NULL, t->ai_addr, RESOLVE_TIMEOUT_MS);
        if (!err)
            break;
    }
    if (err) {
        rdma_error("rdma_resolve_addr() failed with error %d\n", err);
        exit(err);
    }

    /* "get" rdmacm event to acquire event occurred on NIC */
    err = rdma_get_cm_event(client->common.cm_channel, &event);
    if (err) {
        rdma_error("rdma_get_cm_event() failed with error %d\n", err);
        exit(err);
    }

    if (event->event != RDMA_CM_EVENT_ADDR_RESOLVED) {
        rdma_error("rdma_get_cm_event() returned unexpected event %d - expected: RDMA_CM_EVENT_ADDR_RESOLVED\n", event->event);
        exit(1);
    }

    /* Each rdmacm event should be acked */
    rdma_ack_cm_event(event);

    err = rdma_resolve_route(client->common.cm_id, RESOLVE_TIMEOUT_MS);
    if (err) {
        rdma_error("rdma_resolve_route() failed with error %d\n", err);
        exit(err);
    }

    err = rdma_get_cm_event(client->common.cm_channel, &event);
    if (err) {
        rdma_error("rdma_get_cm_event() failed with error %d\n", err);
        exit(err);
    }

    if (event->event != RDMA_CM_EVENT_ROUTE_RESOLVED) {
        rdma_error("rdma_get_cm_event() returned unexpected event %d - expected: RDMA_CM_EVENT_ROUTE_RESOLVED\n", event->event);
        exit(1);
    }

    rdma_ack_cm_event(event);

    err = init_common(&client->common, &config);
    if (err) {
        rdma_error("init_common() failed with error %d\n", err);
        exit(err);
    }

    conn_param.initiator_depth = config.initiator_depth;
    conn_param.responder_resources = config.responder_resources;
    conn_param.retry_count = 7;

	/* Fill the receive queue with empty receive WRs, for synchronizations */
	recv_wr.num_sge = 0;
	for (unsigned i = 1; i < 1024; ++i) {
		err = ibv_post_recv(client->common.cm_id->qp, &recv_wr, &bad_recv_wr);
		if (err) {
			rdma_error("Failed to pre-post the receive buffer, errno: %d \n", err);
			exit(1);
		}
	}

    /* Connect to server */
    err = rdma_connect(client->common.cm_id, &conn_param);
    if (err) {
        rdma_error("rdma_connect() failed with error %d\n", err);
        exit(err);
    }

    err = rdma_get_cm_event(client->common.cm_channel, &event);
    if (err) {
        rdma_error("rdma_get_cm_event() failed with error %d\n", err);
        exit(err);
    }

    if (event->event != RDMA_CM_EVENT_ESTABLISHED) {
        rdma_error("rdma_get_cm_event() returned unexpected event %d - expected: RDMA_CM_EVENT_ESTABLISHED\n", event->event);
        exit(1);
    }

    /* event->param.conn.private_data includes the memory info at server */
    assert(event->param.conn.private_data_len >= sizeof(*client->common.pdata));
    struct pdata *private_pdata_from_server = (struct pdata *)event->param.conn.private_data;
    const size_t expected_private_size = sizeof(*private_pdata_from_server) + private_pdata_from_server->variables * sizeof(private_pdata_from_server->vars[0]);
    assert(event->param.conn.private_data_len >= expected_private_size);
    client->common.pdata = malloc(expected_private_size);
    memcpy(client->common.pdata, event->param.conn.private_data, expected_private_size);

    assert(client->common.pdata->delay_sends.addr);
    assert(client->common.pdata->delay_sends.length == DELAY_SEND_BUF_SIZE);
    assert(client->common.pdata->delay_sends.rkey);

    print_gracefully_rdma_buffer_attr(&client->common);

    rdma_ack_cm_event(event);

    // struct ibv_qp_attr attr = {};
    // struct ibv_qp_init_attr qp_attr = {};
    // err = ibv_query_qp(client->common.cm_id->qp, &attr, IBV_QP_MAX_QP_RD_ATOMIC | IBV_QP_MAX_DEST_RD_ATOMIC, &qp_attr);
    // assert(err == 0);
    // printf("Created QP with max_dest_rd_atomic = %d, max_rd_atomic = %d\n",
    //     attr.max_dest_rd_atomic, attr.max_rd_atomic);

    /* Prepare client's metadata  */
    client->metadata_attr = prepare_common_metadata(&client->common);

    /* now we register the metadata memory */
    client->metadata_mr = ibv_reg_mr(client->common.pd,
                              client->metadata_attr,
                              sizeof(*client->metadata_attr) + sizeof(struct remote_memory_metadata) * client->metadata_attr->variables,
                              IBV_ACCESS_LOCAL_WRITE);

    if (!client->metadata_mr) {
        rdma_error("Failed to register the client metadata buffer\n");
        exit(1);
    }

    /* Send client's metadata, to the server.
     * Fill up SGE */
    sge.addr = (uint64_t)client->metadata_mr->addr;
    sge.length = (uint32_t)client->metadata_mr->length;
    sge.lkey = client->metadata_mr->lkey;
    /* Link to the send work request */
    bzero(&send_wr, sizeof(send_wr));
    send_wr.sg_list = &sge;
    send_wr.num_sge = 1;
    send_wr.opcode = IBV_WR_SEND;
    send_wr.send_flags = IBV_SEND_SIGNALED;

    /* Post it */
    if (ibv_post_send(client->common.cm_id->qp, &send_wr, &bad_send_wr)) {
        rdma_error("Failed to send client metadata, errno: %d \n",
                   -errno);
        exit(-errno);
    }


    /* At this point we are expecting a work completion: the send above */
    int ret = handle_work_completion_status_events(client->common.send_cq, 1);
    if (ret != 1) {
        rdma_error("Failed to get 1 work completion , ret = %d \n", ret);
        exit(ret);
    }
}
