// rdma_setup_server.c

#include "rdma_common_server.h"
#include "handle_work_completion.h"

#include <assert.h>

void rdma_setup_server(struct server_common *common, struct common_setup_config config) {
	struct pdata    				*rep_pdata;
   	struct sockaddr_in     			s_in;
   	struct rdma_cm_event			*event;
   	struct ibv_sge					sge;
   	struct ibv_recv_wr				recv_wr = {}, *bad_recv_wr;
	struct rdma_conn_param conn_param;

    /* Set up RDMA CM structures */
    if (init_common_channel(&common->common))
		exit(1);

    int err = rdma_create_id(common->common.cm_channel, &common->listen_id, NULL, RDMA_PS_TCP); 
    if (err) {
		rdma_error("rdma_create_id() failed: %d\n", err);
        exit(1);
	}

    s_in.sin_family = AF_INET; 
    s_in.sin_port = htons(config.port);
    s_in.sin_addr.s_addr = INADDR_ANY;

    
    /* Bind to local port and listen for connection request */

    err = rdma_bind_addr(common->listen_id, (struct sockaddr *) &s_in);
    if (err) {
		rdma_error("rdma_bind_addr() failed: %d\n", err);
        exit(1);
	}

	/* Now we start to listen on the passed IP */
    err = rdma_listen(common->listen_id, 8); /* backlog = 8 clients */
    if (err) {
		rdma_error("rdma_listen() failed: %d\n", err);
        exit(1);
	}
	
	printf("\n======== server side [%s]===========================\n", config.iteration);
	printf("server is listening for incoming connections...\n");
	fflush(stdout);

	/* now, we expect a client to connect and generate a RDMA_CM_EVNET_CONNECT_REQUEST */
	err = rdma_get_cm_event(common->common.cm_channel, &event);
	if (err) {
		rdma_error("rdma_get_cm_event() failed: %d\n", err);
		exit(1);
	}

	if (event->event != RDMA_CM_EVENT_CONNECT_REQUEST) {
		rdma_error("unexpected event %d\n", event->event);
		exit(1);
	}

	common->common.cm_id = event->id;

	rdma_ack_cm_event(event);

	err = init_common(&common->common, &config);
    if (err) {
		rdma_error("init_common() failed: %d\n", err);
        exit(err);
	}

	const size_t pdata_size = sizeof(struct pdata) + sizeof(struct remote_memory_metadata) * MAX_VARIABLES;
	common->common.pdata = malloc(pdata_size);
	assert(common->common.pdata);
 
	/* we prepare the receive buffer in which we will receive the client metadata */
    common->client_metadata_mr = ibv_reg_mr(common->common.pd,
		common->common.pdata, pdata_size, IBV_ACCESS_LOCAL_WRITE);
	if(!common->client_metadata_mr){
		rdma_error("Failed to register client attr buffer\n");
		//we assume ENOMEM
		exit(1);
	}
	/* We pre-post this receive buffer on the QP. SGE credentials is where we 
	 * receive the metadata from the client */
	sge.addr 	= (uint64_t) common->client_metadata_mr->addr; 
	sge.length 	= common->client_metadata_mr->length;
	sge.lkey 	= common->client_metadata_mr->lkey;
	/* Now we link this SGE to the work request (WR) */
	bzero(&recv_wr, sizeof(recv_wr));
	recv_wr.sg_list = &sge;
	recv_wr.num_sge = 1; // only one SGE
	
	int ret =  ibv_post_recv(common->common.cm_id->qp, &recv_wr, &bad_recv_wr);
	if (ret) {
		rdma_error("Failed to pre-post the receive buffer, errno: %d \n", ret);
		exit(1);
	}

	/* Fill the rest of the receive queue with empty receive WRs, for synchronizations */
	recv_wr.num_sge = 0;
	for (unsigned i = 1; i < 1024; ++i) {
		ret = ibv_post_recv(common->common.cm_id->qp, &recv_wr, &bad_recv_wr);
		if (ret) {
			rdma_error("Failed to pre-post the receive buffer, errno: %d \n", ret);
			exit(1);
		}
	}

		rep_pdata = prepare_common_metadata(&common->common);
        
	    memset(&conn_param, 0, sizeof(conn_param));
	    conn_param.initiator_depth = config.initiator_depth;
	    conn_param.responder_resources = config.responder_resources;  
        conn_param.private_data = rep_pdata; 
        conn_param.private_data_len = sizeof(*rep_pdata) + sizeof(struct remote_memory_metadata) * config.variables;



        /* Accept connection, at this point, server will send the mr info and buffer addr to client to let client directly write data to it,
         * our example here is rep_pdata and conn_param
         */
        err = rdma_accept(common->common.cm_id, &conn_param); 
        if (err) {
            rdma_error("Failed to accept the connection, errno: %d \n", -errno);
	    	exit(1);
        } 
        
		free(rep_pdata);
		rep_pdata = NULL;
        
        err = rdma_get_cm_event(common->common.cm_channel, &event);
        if (err) {
			rdma_error("Failed to retrieve a cm event, errno: %d \n", -errno);
			exit(1);
		}

               
        
        
        /* We expect an RDMA_CM_EVENT_ESTABLISHED to indicate that the RDMA  
		 * connection has been established and everything is fine on both, server 
		 * as well as the client sides.
		 */
        if (event->event != RDMA_CM_EVENT_ESTABLISHED) {
			rdma_error("unexpected event %d\n", event->event);
            exit(1);
		}
        
		// struct ibv_qp_attr attr = {};
		// struct ibv_qp_init_attr qp_attr = {};
		// err = ibv_query_qp(common->common.cm_id->qp, &attr, IBV_QP_MAX_QP_RD_ATOMIC | IBV_QP_MAX_DEST_RD_ATOMIC, &qp_attr);
		// assert(err == 0);
		// printf("Created QP with max_dest_rd_atomic = %d, max_rd_atomic = %d\n",
        // 	attr.max_dest_rd_atomic, attr.max_rd_atomic);
        
        /* Acknowledge the event */
		ret = rdma_ack_cm_event(event);
		if (ret) {
			rdma_error("Failed to acknowledge the cm event %d\n", -errno);
			exit(1);
		}
		
	
	/* Waiting for the client to start the communication by sending to server its
	 * metadata info 
	 */	
	ret = handle_work_completion_status_events(common->common.recv_cq, 1);
	if (ret != 1) {
		rdma_error("Failed to receive, ret = %d \n", ret);
		exit(1);
	}

	debug("Receive buffer pre-posting is successful \n");

	print_gracefully_rdma_buffer_attr(&common->common);
}
