#!/bin/bash -e

CONFIG_FILENAME=config.sh
dir=$(dirname $0)

# Parse command line with getopt
while getopts "i:c:" opt; do
    case $opt in
        i)
            CLIENT_ARGUMENT=$OPTARG
            ;;
        c)
            CONFIG_FILENAME=$OPTARG
            ;;
        \?)
            echo "Invalid option: -$OPTARG" >&2
            exit 1
    esac
done

if [[ ! -e "$CONFIG_FILENAME" ]] ; then
    if [[ -e "$dir/$CONFIG_FILENAME" ]] ; then
        CONFIG_FILENAME="$dir/$CONFIG_FILENAME"
    else 
        echo "Configuration file $CONFIG_FILENAME not found" >&2
        exit 1
    fi
fi

echo Loading configuration file "$CONFIG_FILENAME"
. "$CONFIG_FILENAME"

function run_background_server() {
    local server_program="$1"
    local server_output_file="$2"
    local output_pid_varname="$3"
    shift 3
    local server_args=("$@")

    local sshpass=""
    if [[ -n "$PASSWORD" ]] ; then
        sshpass="sshpass -p '$PASSWORD'"
    fi

	$sshpass ssh "$USER@${SERVER_IDS[0]}" "cd $SERVER_PROGRAM_PATH && ./$server_program ${server_args[@]}" > "$server_output_file" &	
	local pid=$! # fetches the PID of the last background process

	# wait for the server to start listening
	if ! timeout 1 tail -f -n+1 $server_output_file | grep -m1 listening > /dev/null ; then
        kill $pid
        exit 1
    fi

    eval "${output_pid_varname}=$pid"
}

function search_all_strings() {
    local file="$1"
    shift
    local search_strings=("$@")

    local temp_file=$(mktemp /tmp/combined-XXXXX.txt)

    cat "$file" > "$temp_file"

    for search_string in "${search_strings[@]}" ; do
        if ! grep -q "${search_string}" "$temp_file" ; then
            rm "$temp_file"
            return 1
        fi
    done

    rm "$temp_file"

    return 0
}

function run_once() {
    local test="$1"
    local iteration="$2"
    local combined_output_file="$3"
    shift 3
    local search_strings=("$@")

    # Set the path and filenames for the output
    local output_dir=$(pwd)
    local client_output_file="$(mktemp /tmp/client-${test}-XXXXX.txt)"
    local server_output_file="$(mktemp /tmp/server-${test}-XXXXX.txt)"

    local SERVER_PROGRAM="${test}_server"
    local CLIENT_PROGRAM="${test}_client"

	echo "Calling server logic..."
    run_background_server "$SERVER_PROGRAM" "$server_output_file" p1_pid "$iteration"
	
	echo "Calling client logic..."
	./$CLIENT_PROGRAM ${SERVER_IPS[0]} $i > "$client_output_file"
	
    wait $p1_pid

    cat "$client_output_file" "$server_output_file" >> "$combined_output_file"

    search_all_strings <(cat "$client_output_file" "$server_output_file") "${search_strings[@]}" # return code as result of this function
}

function run_two_servers_once() {
    local test="$1"
    local iteration="$2"
    local combined_output_file="$3"
    shift 3
    local search_strings=("$@")

    # Set the path and filenames for the output
    local output_dir=$(pwd)
    local client_output_file="$(mktemp /tmp/client-${test}-XXXXX.txt)"
    local server1_output_file="$(mktemp /tmp/server-${test}-XXXXX.txt)"
    local server2_output_file="$(mktemp /tmp/server-${test}-XXXXX.txt)"

    local SERVER_PROGRAM="${test}_server"
    local CLIENT_PROGRAM="${test}_client"

	echo "Calling server logic..."
    run_background_server "$SERVER_PROGRAM" "$server1_output_file" server1_pid 1 "$iteration"
    run_background_server "$SERVER_PROGRAM" "$server2_output_file" server2_pid 2 "$iteration-2"
	
	echo "Calling client logic..."
	./$CLIENT_PROGRAM ${SERVER_IPS[0]} ${SERVER_IPS[1]} $i > "$client_output_file"
	
    wait $server1_pid $server2_pid

    cat "$client_output_file" "$server1_output_file" "$server2_output_file" >> "$combined_output_file"

    search_all_strings <(cat "$client_output_file" "$server1_output_file" "$server2_output_file") "${search_strings[@]}" # return code as result of this function
}

function join_by {
  local d=${1-} f=${2-}
  if shift 2; then
    printf %s "$f" "${@/#/$d}"
  fi
}

function run_multiple_func() {
    local test="$1"
    local func="$2"
    shift 2
    local search_strings=("$@")

    local search_string_printable="$(join_by " and " "${search_strings[@]}")"

    local combined_output_file="output_${test}_$(date +'%Y%m%d_%H%M%S').txt"

    # Initialize counter
    local -i search_string_count=0

    # Initialize array to store iterations where search string was found
    search_string_iterations=()

    for ((i=1; i<=$NUM_ITERATIONS; i++))
    do
        echo -e "\n\nIteration $i"
        # Run once and check if search string was found
        if "$func" "$test" "$i" "$combined_output_file" "${search_strings[@]}" ; then
            let search_string_count++ || true
            # Add the iteration to the array
            search_string_iterations+=("$i")
        fi

        # Calculate percentage of occurrences
        local percentage=$(echo "scale=2; (${search_string_count} / $i) * 100" | bc)

        # Print occurrences and counter on each iteration
        echo -e "\nOccurrences of '$search_string_printable' on iteration $i: $search_string_count [$percentage%]"
    done

    cat "$combined_output_file"

    final_percentage=$(echo "scale=2; ($search_string_count / $NUM_ITERATIONS) * 100" | bc)

    echo -e "\n\n________________________________________"
    echo -e "\nNumber of iterations with both '$search_string_printable' happening in the same iteration: $search_string_count [$final_percentage%]"
    echo -e "\n________________________________________"

    # Print list of iterations where $search_string
    echo -e "\nIterations where $search_string_printable was observed: ${search_string_iterations[@]}"

    echo -e "\n\n"
}

function run_multiple() {
    local test="$1"
    shift
    local search_strings=("$@")

    run_multiple_func "$test" "run_once" "${search_strings[@]}"
}