/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make rff1_client
 *** gcc -g -o rff1_client rff1_client.c rdma_setup_client.c handle_work_completion.c disconnect_cleanup.c -lrdmacm -libverbs
 *
 * usage：
 * ./rff1_client 192.168.0.2 <index> // to sgx02
 * 
 *     P1        │   P2
 *     init z=0  │  init x,y=0
 *     ──────────┤─────────────
 *     z  := x²  │  y := z¹
 *     a  := y²  │  poll(1)
 *     y² := 2   │  b := y
 *     x² := 1   |
 *     ─────────────────────────
 *     (a,b)=(1,1) possible, (1,2) not possible
 *
 *
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client [local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ z ... buf[0]   │<── z:=x² ────│ x ... buf[0]   │
 *    │                │─── x²:=1 ───>│                │
 *    │                │              ├────────────────┤ 
 *    │                │─── y:=z¹ ───>│ y ... buf[1]   │
 *    │                │<── y²:=2 ────│                │
 *    ├────────────────┤              │                │
 *    │ a ... buf[1]   │<── a:=y² ────│                │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"

#include <assert.h>

int main(int argc, char *argv[]) {

	char *itr = (argc > 2) ? argv[2] : "(null)";	
    // Seed the random number generator
    srand(time(0));
	struct client_common client = {};

/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */

    struct common_setup_config config = default_config(3 /* variables */, itr);
    // config.initiator_depth = 1;
    rdma_setup_client(&client, argv[1], config);
   
    // Define pointer aliases
    const local_var z    = {0},
                    a    = {1},
                    sync = {2};
    const remote_var x = {0},
                     y = {1};

// Begin of Litmus test

    enum { num_ops = 5, };
    struct ibv_send_wr send_wr[num_ops] = {}, *bad_wr;
    struct ibv_sge sge[num_ops] = {};

    for (int i = 0; i < num_ops - 1; i++)
        send_wr[i].next = &send_wr[i + 1];
    for (int i = 0; i < num_ops; i++) {
        sge[i].length = sizeof(uint32_t);
        send_wr[i].sg_list = &sge[i];
        send_wr[i].num_sge = 1;
        send_wr[i].send_flags = IBV_SEND_SIGNALED;
        send_wr[i].wr_id = i;
    }

    int i = -1;

	/* z  := x²  */
    ++i;
    set_local_var(&client.common, &sge[i], z);
    set_remote_var(&client.common, &send_wr[i], x);
    send_wr[i].opcode = IBV_WR_RDMA_READ;

    /* read for delay */
    ++i;
    sge[i].addr = (uintptr_t)client.common.delay_sends_buf;
	sge[i].length = DELAY_SEND_BUF_SIZE - ((rand() % 4)  * 64);
	sge[i].lkey = client.common.delay_sends_mr->lkey;
	send_wr[i].wr_id = RANDOM_NIC_DELAY_WR_ID + 1;
	send_wr[i].opcode = IBV_WR_RDMA_READ;
	send_wr[i].send_flags = 0; // no CQE unless there is an error
	send_wr[i].wr.rdma.rkey = ntohl(client.common.pdata->delay_sends.rkey);
	send_wr[i].wr.rdma.remote_addr = client.common.pdata->delay_sends.addr;

    /* a  := y²  */
    ++i;
    set_local_var(&client.common, &sge[i], a);
    set_remote_var(&client.common, &send_wr[i], y);
    send_wr[i].opcode = IBV_WR_RDMA_READ;

	/* y² := 2 */
    ++i;
    uint32_t value0 = 2;
    sge[i].addr = (uint64_t)(uintptr_t)&value0;
    set_remote_var(&client.common, &send_wr[i], y);
    send_wr[i].opcode = IBV_WR_RDMA_WRITE;
    send_wr[i].send_flags |= IBV_SEND_INLINE;

    /* x² := 1 */
    ++i;
    uint32_t value1 = 1;
    sge[i].addr = (uint64_t)(uintptr_t)&value1;
    set_remote_var(&client.common, &send_wr[i], x);
    send_wr[i].opcode = IBV_WR_RDMA_WRITE;
    send_wr[i].send_flags |= IBV_SEND_INLINE;

    int ret = ibv_post_send(client.common.cm_id->qp, send_wr, &bad_wr);
    assert(ret == 0);

	/* Wait for the server to indicate the completion of the operation  */
	ret = handle_work_completion_status_events(client.common.send_cq, 4);
	if (ret != 4) {
		rdma_error("Failed to get 4 work completion from server, ret = %d \n", ret);
		return ret;
	}

// End of Litmus test

    // Wait for sync (write with immediate)
	ret = handle_work_completion_status_events(client.common.recv_cq, 1);
	if (ret != 1) {
		rdma_error("Failed to get a work completion from server, ret = %d \n", ret);
		return ret;
	}

    uint32_t b = load_local(&client.common, sync);

    // Call the disconnect() function
    int result = disconnect(&client);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

	printf("Upon Litmus Test execution: (a,b) == (%u,%u) [z == %u]\n",
        load_local(&client.common, a), b, load_local(&client.common, z));

    // Call the cleanup function
    result = cleanup(&client);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }    

    return 0;
}



	
	

