/* 
 *  █▀ █▀▀ █▀█ █░█ █▀▀ █▀█ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  ▄█ ██▄ █▀▄ ▀▄▀ ██▄ █▀▄ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build:
 * gcc -o rff1_server rff1_server.c -lrdmacm -libverbs
 * 
 * usage: 
 * ./rff1_server <index>
 * 
 *     P1        │   P2
 *     init z=0  │  init x,y=0
 *     ──────────┤─────────────
 *     z  := x²  │  y := z¹
 *     a  := y²  │  poll(1)
 *     y² := 2   │  b := y
 *     x² := 1   |
 *     ─────────────────────────
 *     (a,b)=(1,1) possible, (1,2) not possible
 *
 *
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client [local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ z ... buf[0]   │<── z:=x² ────│ x ... buf[0]   │
 *    │                │─── x²:=1 ───>│                │
 *    │                │              ├────────────────┤ 
 *    │                │─── y:=z¹ ───>│ y ... buf[1]   │
 *    │                │<── y²:=2 ────│                │
 *    ├────────────────┤              │                │
 *    │ a ... buf[1]   │<── a:=y² ────│                │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_server.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_server.h"
#include "rdma_setup_server.h"
#include "read.h"
#include "write.h"

#include <assert.h>

int main(int argc, char *argv[]) 
{ 
	char *iter = (argc > 1) ? argv[1] : "(null)";
	
    // Seed the random number generator
    srand(time(0));

	struct server_common server = {};

/*
 - Setup RDMA resources
 - Exchange information about the local buffer between the client and server 
 */
	rdma_setup_server(&server, default_config(2 /* variables */, iter));

    // Define pointer aliases
    const local_var x = {0},
                    y = {1};
    const remote_var z    = {0},
                     sync = {2};

    (void)x;
/*
 ◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤
 ◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣
	
  _     _ _                         _____         _   
 | |   (_) |_ _ __ ___  _   _ ___  |_   _|__  ___| |_ 
 | |   | | __| '_ ` _ \| | | / __|   | |/ _ \/ __| __|
 | |___| | |_| | | | | | |_| \__ \   | |  __/\__ \ |_ 
 |_____|_|\__|_| |_| |_|\__,_|___/   |_|\___||___/\__|

*/

    // random_cpu_delay();
    // random_nic_delay(&server.common);
    // random_nic_delay(&server.common);

	/* y := z¹  */
    read_to_from(&server.common, y, z);

	/* Wait for the client to indicate the completion of the operation  */
	int ret = handle_work_completion_status_events(server.common.send_cq, 1);
	if (ret != 1) {
		rdma_error("Failed to get 1 work completion from client, ret = %d \n", ret);
		return ret;
	}

    uint32_t b = load_local(&server.common, y);

/*
 ◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤
 ◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣◥◤◢◣
*/ 
    write_inline_w_imm(&server.common, sync, b, 0x80000000ul, true /* fence */);

    // Call the disconnect() function
    int result = disconnect(&server);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

	printf("Upon Litmus Test execution: b == %u, [x == %u, y == %u]\n",
        b, load_local(&server.common, x), load_local(&server.common, y));


    // Call the cleanup function
    result = cleanup(&server);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }

		
    return 0;
}
