/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make rrpr3_client
 *** gcc -g -o rrpr3_client rrpr3_client.c rdma_setup_client.c handle_work_completion.c disconnect_cleanup.c -lrdmacm -libverbs
 *
 * usage：
 * ./rrpr3_client 192.168.0.2 <index> // to sgx02
 * 
 *     P1        │   P2
 *     init x,y=1│  init z=0
 *     ──────────┤─────────────
 *     x  := z²  │  
 *     y  := z²  │ 
 *     a  := y   │  
 *     b  := x   │  
 *     ─────────────────────────
 *     (a,b)=(0,1) not possible
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client (local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ x ... buf[0]   │<── x:=z² ────│                │
 *    ├────────────────┤              │ z ... buf[0]   │ 
 *    │ y ... buf[1]   │<── y:=z² ────│                │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"

#include <assert.h>

int main(int argc, char *argv[]) {

	char *itr = (argc > 2) ? argv[2] : "(null)";	
    // Seed the random number generator
    srand(time(0));
	struct client_common client = {};
    int ret;

/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */

    struct common_setup_config config = default_config(3 /* variables */, itr);
    config.randomize_variable_home = true;
    rdma_setup_client(&client, argv[1], config);
   
    // Define pointer aliases
    const local_var x    = {0}, 
    				y    = {1};
    const remote_var z = {0};

	store_local(&client.common, x, 1);
	store_local(&client.common, y, 1);
	
// Begin of Litmus test
    /* x  := z²  */
    read_to_from(&client.common, x, z);

    /* y  := z²  */
    read_to_from(&client.common, y, z);

	uint32_t a = load_local(&client.common, y);
	
	uint32_t b = load_local(&client.common, x);


    /* Wait for the client to indicate the completion of the operation  */
    ret = handle_work_completion_status_events(client.common.send_cq, 2);
    if (ret != 2) {
        rdma_error("Failed to get 2 work completions from server, ret = %d \n", ret);
        return ret;
    }

// End of Litmus test

  
    // Call the disconnect() function
    int result = disconnect(&client);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

	printf("[%s] Upon Litmus Test execution: (a,b) == (%u,%u)\n",
        itr, a, b);

    // Call the cleanup function
    result = cleanup(&client);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }    

    return 0;
}



	
	

