/* 
 *  █▀▀ █░░ █ █▀▀ █▄░█ ▀█▀ ▄▄ █▀ █ █▀▄ █▀▀   █░░ █▀█ █▀▀ █ █▀▀
 *  █▄▄ █▄▄ █ ██▄ █░▀█ ░█░ ░░ ▄█ █ █▄▀ ██▄   █▄▄ █▄█ █▄█ █ █▄▄
 *
 *
 * build：
 *   make rrr_client
 *** gcc -g -o rrr_client rrr_client.c rdma_setup_client.c handle_work_completion.c disconnect_cleanup.c -lrdmacm -libverbs
 *
 * usage：
 * ./rrr_client 192.168.0.2 <index> // to sgx02
 * 
 *     P1        │   P2
 *               │  init x=0
 *     ──────────┤─────────────
 *     a  := x²  │  x  := 1
 *     b  := x²  │  x  := 2
 *     ─────────────────────────
 *     (a,b)=(2,1) possible
 *
 *    ╔════════════════╗              ╔════════════════╗
 *    ║ Client (local) ║              ║ Server (remote)║
 *    ╚════════════════╝              ╚════════════════╝
 *    │ a ... buf[0]   │<── a:=x² ────│ x ... buf[0]   │
 *    ├────────────────┤              │                │ 
 *    │ b ... buf[1]   │<── b:=x² ────│                │
 *    └────────────────┘              └────────────────┘
 */
#include "common.h"
#include "rdma_common_client.h"
#include "handle_work_completion.h"
#include "disconnect_cleanup_client.h"
#include "rdma_setup_client.h"
#include "read.h"
#include "write.h"

#include <assert.h>

int main(int argc, char *argv[]) {

	char *itr = (argc > 2) ? argv[2] : "(null)";	
    // Seed the random number generator
    srand(time(0));
	struct client_common client = {};
    int ret;

/*
 * Setup RDMA resources
 * Exchange information about the local buffer between the client and server 
 */

    struct common_setup_config config = default_config(3 /* variables */, itr);
    config.randomize_variable_home = true;
    rdma_setup_client(&client, argv[1], config);
   
    // Define pointer aliases
    const local_var a    = {0}, 
    				b    = {1},
                    sync = {2};
    const remote_var x = {0};

    (void)sync;

// Begin of Litmus test
    /* a  := x²  */
    read_to_from(&client.common, a, x);

    /* b  := x²  */
    read_to_from(&client.common, b, x);

    /* Wait for the client to indicate the completion of the operation  */
    ret = handle_work_completion_status_events(client.common.send_cq, 2);
    if (ret != 2) {
        rdma_error("Failed to get 2 work completions from server, ret = %d \n", ret);
        return ret;
    }

// End of Litmus test

    /* Wait for the sync message */
	ret = handle_work_completion_status_events(client.common.recv_cq, 1);
    assert(ret == 1);
  
    // Call the disconnect() function
    int result = disconnect(&client);
    if (result != 0) {
        fprintf(stderr, "Error: disconnect failed with error code %d\n", result);
    }

	printf("[%s] Upon Litmus Test execution: (a,b) == (%u,%u)\n",
        itr, load_local(&client.common, a), load_local(&client.common, b));

    // Call the cleanup function
    result = cleanup(&client);
    if (result != 0) {
        fprintf(stderr, "Error: cleanup failed with error code %d\n", result);
    }    

    return 0;
}



	
	

